@@ -84,13 +84,13 @@ sub install {
 }
 
 sub cmd_export {
-    my($self, @args) = @_;
+    my($self) = @_;
     my %env = $self->env;
     print "export PATH=$env{PATH} PERL5LIB=$env{PERL5LIB}\n";
 }
 
 sub cmd_env {
-    my($self, @args) = @_;
+    my($self) = @_;
     my %env = $self->env;
     print "PATH=$env{PATH}\nPERL5LIB=$env{PERL5LIB}\n";
 }
@@ -114,11 +114,24 @@ sub cmd_find {
 sub cmd_list {
     my($self) = @_;
 
-    for my $artifact ($self->resolve) {
+    my @artifacts;
+    $self->resolve(sub { push @artifacts, $_[0] });
+
+    for my $artifact (sort { $a->package cmp $b->package } @artifacts) {
         printf "%s (%s) in %s\n", $artifact->package, $artifact->version || '0', $artifact->path;
     }
 }
 
+sub cmd_tree {
+    my($self) = @_;
+
+    $self->resolve(sub {
+        my($artifact, $depth) = @_;
+        printf "%s%s (%s) in %s\n", (" " x $depth), $artifact->package, $artifact->version || '0', $artifact->path;
+    });
+
+}
+
 sub cmd_index {
     my $self = shift;
     $self->write_index(*STDOUT);
@@ -128,7 +141,8 @@ sub write_index {
     my($self, $fh) = @_;
 
     my @packages;
-    for my $artifact ($self->resolve) {
+    $self->resolve(sub {
+        my $artifact = shift;
         while (my($package, $data) = each %{$artifact->install->{provides}}) {
             push @packages, {
                 package => $package,
@@ -136,7 +150,7 @@ sub write_index {
                 pathname => $artifact->install->{pathname},
             }
         }
-    }
+    });
 
     print $fh <<EOF;
 File:         02packages.details.txt
@@ -217,12 +231,10 @@ sub snapshot_to_requirements {
 }
 
 sub resolve_recursive {
-    my($self, $requirements, $seen, $cb, $missing_cb) = @_;
+    my($self, $root_reqs, $requirements, $seen, $cb, $depth) = @_;
 
     my $repo = $self->build_repo;
-    my $recurse;
 
-    my @missing;
     for my $module (sort $requirements->required_modules) {
         next if $module eq 'perl';
 
@@ -231,44 +243,34 @@ sub resolve_recursive {
             next;
         } elsif (my $artifact = $repo->find($module, $want_version)) {
             next if $seen->{$artifact->path}++;
-            $cb->($artifact);
+            $cb->($artifact, $depth);
             my $meta = CPAN::Meta->load_file($artifact->path . "/MYMETA.json");
-            $requirements->add_requirements($meta->effective_prereqs->merged_requirements(['runtime'], ['requires']));
+            my $reqs = $meta->effective_prereqs->merged_requirements(['runtime'], ['requires']);
+            $root_reqs->add_requirements($reqs);
+            $self->resolve_recursive($root_reqs, $reqs, $seen, $cb, $depth + 1);
         } else {
-            push @missing, [ $module, $want_version ];
+            die "Coulld not find an artifact for $module => $want_version\n";
         }
-
-        $recurse++;
     }
-
-    $missing_cb->(@missing) if @missing;
-    $self->resolve_recursive($requirements, $seen, $cb, $missing_cb) if $recurse;
 }
 
 sub resolve {
-    my($self, $requirements) = @_;
+    my($self, $cb) = @_;
 
-    $requirements ||= $self->build_requirements
+    my $requirements = $self->build_requirements
       or Carp::croak "Could not locate 'cpanfile' to load module list.";
 
     my @artifacts;
-    $self->resolve_recursive($requirements, {}, sub { push @artifacts, @_ }, sub { $self->warn_missing(@_) });
+    $self->resolve_recursive($requirements, $requirements, {}, $cb, 0);
 
     @artifacts;
 }
 
-sub warn_missing {
-    my($self, @missing) = @_;
-
-    for my $missing (@missing) {
-        Carp::carp "Could not find an artifact for $missing->[0] => $missing->[1]";
-    }
-}
-
 sub env {
-    my($self, @args) = @_;
+    my($self) = @_;
 
-    my @artifacts = $self->resolve(@args);
+    my @artifacts;
+    $self->resolve(sub { push @artifacts, $_[0] });
     return (
         _join(PATH => map $_->paths, @artifacts),
         _join(PERL5LIB => map $_->libs, @artifacts),
